#!/usr/bin/env node

import { hideBin } from "yargs/helpers";
import yargs from "yargs";
import { join } from "path";
import yaml from "yaml";
import fs from "fs-extra";
import consola from "consola";
import u from "@yanick/updeep-remeda";
import { once } from "remeda";
import simpleGit from "simple-git";
import Ajv from "ajv";
import { nanoid } from "nanoid";
import slugify from "@sindresorhus/slugify";

import print from "./command/print.js";
import init from "./command/init.js";
import schema from "./command/schema.js";
import add from "./command/add.js";
import cut from "./command/cut.js";
import upcoming, { next_release } from "./command/upcoming.js";
import latest, { latest_version } from "./command/latest-version.js";
import validate from "./command/validate.js";
import git_gather from "./command/git-gather.js";
import schemaV1 from "./changelog-schema.js";
import { globby } from "globby";
import { flatMap } from "remeda";

consola.raw = (...args) => console.log(...args);

yargs(hideBin(process.argv))
  .middleware((config) => {
    config.git = once(simpleGit);
    config.consola = consola;
    config.changelog = once(() =>
      fs
        .readFile(config.source, "utf-8")
        .then(yaml.parse)
        .then(async (doc) => {
          if (!doc.project.next_directory) return doc;

          const changes = await globby([
            doc.project.next_directory + "/*.yml",
            doc.project.next_directory + "/*.yaml",
          ])
            .then((files) =>
              Promise.all(
                files.map((f) => fs.readFile(f, "utf-8").then(yaml.parse))
              )
            )
            .then((r) => r.flat());

          if (changes.length)
            doc.releases
              .find((r) => r.version === "NEXT")
              .changes.push(...changes);

          return doc;
        })
        .then((doc) => {
          const ajv = new Ajv();
          const validate = ajv.compile(schemaV1);
          const valid = validate(doc);
          if (valid) return doc;

          config.consola.error("invalid changelog: ", validate.errors);
          throw "changelog is invalid";
        })
    );
    config.delete_next_dir_entries = async () => {
      const changelog = await config.changelog();
      return globby([
        changelog.project.next_directory + "/*.yml",
        changelog.project.next_directory + "/*.yaml",
      ]).then((files) => Promise.all(files.map((f) => fs.remove(f))));
    };
    config.save_changelog = async (changelog) => {
      if (!changelog) changelog = await config.changelog();
      return fs.writeFile(config.source, yaml.stringify(changelog));
    };
    config.next_release = next_release.bind(config);
    config.latest_version = latest_version.bind(config);
    return config;
  })
  .middleware((argv, yargs) => {
    argv.yargs = yargs;

    argv.add_to_next = async (entry) => {
      const dir = await argv.changelog().then((c) => c.project.next_directory);

      if (dir) {
        await fs.ensureDir(dir);
        const filename = join(
          dir,
          [
            new Date().toISOString(),
            entry.ticket,
            entry.feat,
            slugify(entry.desc, {
              separator: "_",
            }).slice(0, 10),
          ]
            .filter((x) => x)
            .join("-") + ".yml"
        );
        argv.consola.info(`writing change to ${filename}`);
        if (fs.existsSync(filename)) {
          argv.consola.error(`file ${filename} already exist`);
          yargs.exit(1);
        }
        if (Object.keys(entry).length === 1) {
          entry = entry.desc;
        }
        return fs.writeFile(filename, yaml.stringify([entry]));
      }

      const changelog = await argv.changelog();

      let next = changelog.releases.find(u.matches({ version: "NEXT" }));
      if (!next) {
        changelog.releases.unshift(base_next_version);
      }

      if (Object.keys(entry).length === 1) {
        entry = entry.desc;
      }

      next.changes.push(entry);

      return argv.save_changelog();
    };
  })
  .default("source", join(process.cwd(), "CHANGELOG.yml"))
  .describe("source", "changelog source")
  .command(init)
  .command(add)
  .command(print)
  .command(cut)
  .command(schema)
  .command(upcoming)
  .command({
    ...upcoming,
    command: "next",
    desc: 'alias for "upcoming"',
  })
  .command(latest)
  .command(git_gather)
  .command(validate)
  .strictCommands()
  .help()
  .parse();
